import typer
import os
from fastmvc import __version__
from fastmvc.generator import build_base, gen_scaffold, gen_authlib
from typing import List
import fastmvc.utilities as utils

app = typer.Typer()


@app.command()
def version():
    """ show current version of FastMVC"""
    typer.echo(f"FastMVC {__version__}")


def __choose_platform(param_platform):
    plats = utils.platforms()
    if param_platform:
        try:
            selected_platform = utils.Platform[param_platform]
        except:
            selected_platform = None
    else:
        typer.secho('\n'.join(f"[{k}] {v.name}" for k, v in plats.items()))
        selected_platform = typer.prompt("Which platform would you like to build for?")
        selected_platform = plats.get(selected_platform)
    if not selected_platform:
        typer.secho(f"Please choose a valid Platform. {list(plats.values())}", fg='red')
    return selected_platform


@app.command()
def new(project: str, platform: str or None = None):
    """ create a new project """
    plat = __choose_platform(platform)
    if not plat:
        return
    
    new_folder = os.path.join(os.curdir, project)
    try:      
        typer.secho(f"\nBuilding new project: {project}\n", fg='green')
        build_base(new_folder, project, plat)
        typer.echo(f"\n{project} was created.\n")
    except FileExistsError:
        typer.secho(f"'{project}' already exists in this folder.\n", fg='red')
    

@app.command()
def server():
    """ run the app locally """
    utils.run_server()


@app.command()
def s():
    """ alias for 'server' """
    server()


@app.command()
def scaffold(obj: str, attributes: List[str]):
    """ create a router and views for a described object """
    typer.secho(f"\nScaffolding views and router for: {obj}\n", fg='green')
    gen_scaffold(os.curdir, obj, attributes)
    typer.echo(f"\n{obj} was created.\n")


@app.command()
def auth():
    """ generate authorization framework for users """
    typer.secho(f"\nGenerating views and router for: User\n", fg='green')
    gen_authlib(os.curdir)
    typer.echo(f"\nAuth was created.\n")


# @app.command()
# def set_project_key(project_key: str):
#     """ set your development project key for Deta for faster builds """
#     utils.set_project_key(project_key)
#     typer.secho("Project Key successfully set.", fg='green')


# @app.command()
# def clear_project_key():
#     """ clear your Deta development project key """
#     utils.clear_project_key()
#     typer.secho("Project Key successfully cleared.", fg='green')


@app.command()
def view_config():
    """ view your development configurations """
    conf = utils.config()
    typer.echo(conf)



from google.cloud import firestore
from fastmvc.models.database._base import BaseDataModel, DBModelMetaClass


class FirestoreModelMeta(DBModelMetaClass):

    @staticmethod
    def handle_db_property(cls):
        if cls._db:
            return cls._db
        cls._db = firestore.Client().collection(cls.__db_name__)
        return cls._db
    

class Firestore(BaseDataModel, metaclass=FirestoreModelMeta):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def __parsed(cls, record):
        r = record.to_dict()
        r.update({'key': record.id})
        return cls.parse_obj(r)

    @classmethod
    def get(cls, key):
        doc_ref = cls.__db__.document(key)
        doc = doc_ref.get()
        if doc.exists:
            return cls.__parsed(doc)
        else:
            return None
    
    @classmethod
    def get_all(cls):
        for record in cls.__db__.get():
            yield cls.__parsed(record)
    
    @classmethod
    def delete_key(cls, key):
        cls.__db__.document(key).delete()

    @classmethod 
    def query(cls, statement: dict):
        query = cls.__db__
        for k, v in statement.items():
            query = query.where(k, '==', v)
        for record in query.stream():
            yield cls.__parsed(record)
            
    @classmethod
    def _edit(cls, key, data):
        doc_ref = cls.__db__.document(key)
        return doc_ref.update(data)

    def update(self, update_data: dict):
        result = self._edit(key=self.key, data=update_data)

    @classmethod 
    def _put(cls, key, data):
        doc_ref = cls.__db__.document(key)
        doc_ref.set(data)
        return doc_ref.id
    
    def save(self):
        data = self.dict()
        key = data.pop('key', None)
        new_doc_id = self._put(key, data)
        self.key = new_doc_id



from datetime import datetime, timedelta
from functools import wraps

import jwt

from authlib.integrations.starlette_client import OAuth
from fastapi import HTTPException, Request
from fastapi.responses import RedirectResponse
from passlib.context import CryptContext
from starlette.config import Config


def current_user(request: Request):
    user = request.session.get('user')
    return user


def authenticated_path(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        user = current_user(kwargs['request'])
        if not user:
            return RedirectResponse('/user/not_authenticated', 303)
        return func(*args, **kwargs)
    return wrapper


class Auth():
    config = Config('.env')
    hasher= CryptContext(schemes=['bcrypt'])
    secret = config.get("APP_SECRET")

    # GOOGLE OAUTH
    """
    Configuring Google Auth2.0 API Client:

    https://developers.google.com/identity/oauth2/web/guides/get-google-api-clientid
    click on 'configure a project'
    """

    oauth = OAuth(config)

    oauth.register(
        name='google',
        server_metadata_url='https://accounts.google.com/.well-known/openid-configuration',
        client_kwargs={
            'scope': 'openid email profile'
        }
    )
    # ============

    def encode_password(self, password):
        return self.hasher.hash(password)

    def verify_password(self, password, encoded_password):
        return self.hasher.verify(password, encoded_password)

    def encode_token(
        self,
        username,
        scope='access_token',
        exp=datetime.utcnow() + timedelta(days=14)):

        payload = {
            'exp' : exp,
            'iat' : datetime.utcnow(),
                'scope': scope,
            'sub' : username
        }

        encoded = jwt.encode(
            payload,
            self.secret,
            algorithm='HS256'
        )
        if isinstance(encoded, bytes):
            encoded = encoded.decode('utf-8')
        return encoded

    def decode_token(self, token):
        try:
            payload = jwt.decode(token, self.secret, algorithms=['HS256'])
            if (payload['scope'] == 'access_token'):
                return payload['sub']
            raise HTTPException(status_code=401, detail='Scope for the token is invalid')
        except jwt.ExpiredSignatureError:
            raise HTTPException(status_code=401, detail='Token expired')
        except jwt.InvalidTokenError:
            raise HTTPException(status_code=401, detail='Invalid token')

    def encode_refresh_token(self, username):
        return self.encode_token(
            username,
            scope='refresh_token',
            exp=datetime.utcnow() + timedelta(days=30))

    def refresh_token(self, refresh_token):
        try:
            payload = jwt.decode(refresh_token, self.secret, algorithms=['HS256'])
            if (payload['scope'] == 'refresh_token'):
                username = payload['sub']
                new_token = self.encode_token(username)
                return new_token
            raise HTTPException(status_code=401, detail='Invalid scope for token')
        except jwt.ExpiredSignatureError:
            raise HTTPException(status_code=401, detail='Refresh token expired')
        except jwt.InvalidTokenError:
            raise HTTPException(status_code=401, detail='Invalid refresh token')

    @classmethod
    def is_google_oauth2_enabled(cls):
        if all([cls.config.get('GOOGLE_CLIENT_ID'),
                cls.config.get('GOOGLE_CLIENT_SECRET')]):
            return True
        return False



from fastapi import Request
from fastapi.responses import RedirectResponse
from fastmvc.models.html import Alert 
from fastmvc.models.database.^{platform.data_model_import}^
from user.utils import Auth, current_user


auth_handler = Auth()


class User(^{platform.data_model}^):
    username: str
    password: str

    class Config:
        table_name = "^{proj}^_user"

    @classmethod
    def fetch_user(cls, username: str):
        user = [u for u in cls.query({'username': username})]
        if user:
            return user[0]
        return False

    def signup(self):
        if User.fetch_user(self.username):
            return Alert('An account with that username already exists.')
        else:
            try:
                hashed_password = auth_handler.encode_password(self.password)
                self.password = hashed_password
                self.save()
                return Alert('Success! Please login.', 'success')
            except:
                return Alert('Failed to signup User.')

    @classmethod
    def new_user(cls, email):
        data = {
            'username': email,
            'password': 'N/A'
        }
        user = User.parse_obj(data)
        user.save()
        return user

    @classmethod
    def signup_from_google_login(cls, email):
        user = cls.fetch_user(email)
        if not user:
            user = cls.new_user(email)
        return user

    def create_session(self, request: Request):
        access_token = auth_handler.encode_token(self.key)
        refresh_token = auth_handler.encode_refresh_token(self.key)
        results = {
            'access_token': access_token,
            'refresh_token': refresh_token}

        request.session['user'] = results
        request.session['user'].update(self.dict())

    def login(self, request: Request):
        find_user = [u for u in User.query({'username': self.username})]
        alert = Alert('Invalid Username or Password')
        if find_user:
            user = find_user[0]
        else:
            return alert
        if not auth_handler.verify_password(self.password, user.password):
            return alert

        user.create_session(request)
        return RedirectResponse('/user/dashboard', 303)

    @classmethod
    def refresh_token(cls, request: Request):
        user = current_user(request)
        refresh_token = user['refresh_token']
        return auth_handler.refresh_token(refresh_token)


import json 
from pathlib import Path
import os
import enum

CONFIG_PATH = Path(__file__).parent.resolve() / 'user_config.json'


class Platform(enum.Enum):
    GOOGLE_APP_ENGINE = 'GOOGLE_APP_ENGINE'
    DETA = 'DETA'


def platforms():
    platforms = dict()
    for i, p in enumerate(Platform):
        platforms[f"{i}"] = p
    return platforms


def get_project_platform():
    for (_, _, filenames) in os.walk(os.curdir):
        for fn in filenames:
            if 'app.yaml' in fn:
                return Platform.GOOGLE_APP_ENGINE
    return Platform.DETA


def __fetch_config():
    with open(CONFIG_PATH, 'r') as j:
        return json.loads(j.read())


def set_project_key(project_key: str):
    conf = __fetch_config()
    conf['project_key'] = project_key 
    with open(CONFIG_PATH, 'w') as j:
        j.write(json.dumps(conf, indent=4))


def clear_project_key():
    set_project_key("")


def config(key: str or None = None):
    if key:
        return __fetch_config().get(key)
    return __fetch_config()


def run_server():
    server_cmd = "uvicorn main:app --reload"
    os.system(server_cmd)



import os

import flickr_api

    
def setup_flickr_api():
    flickr_api.set_keys(
        api_key=os.environ['FLICKR_API_KEY'], 
        api_secret=os.environ['FLICKR_API_SECRET'])
    auth_handler = flickr_api.auth.AuthHandler(
        access_token_key=os.environ['FLICKR_ACCESS_TOKEN'], 
        access_token_secret=os.environ['FLICKR_ACCESS_SECRET'])
    flickr_api.set_auth_handler(auth_handler)


def __get_sizes(result: flickr_api.Photo):
    new_dict = dict()
    for nm, vals in result.getSizes().items():
        newname = nm.replace(' ', '_').lower()
        new_dict[newname] = vals
    return new_dict


def upload_photo(filepath: str, title: str, file_data=None):
    result = flickr_api.upload(
        photo_file=filepath, 
        photo_file_data=file_data,
        title=title, 
        is_public="0")
    return {
        'flickr_id': result.id, 
        'sizes': __get_sizes(result)}


def delete_photo(id: str):
    flickr_api.Photo(id=id).delete()


def authorize_app(api_key: str, api_secret: str, permission: str = 'read'):
    """
    Run this from the commandline when you need to authorize an app.

    Permission options: 'read', 'write', 'delete'
    """
    a = flickr_api.auth.AuthHandler(key=api_key, secret=api_secret) 
    url = a.get_authorization_url(permission)
    print('navigate to the following url:', url)
    print("Click 'Accept', you will be taken to an xml.")
    print("Locate the field 'oauth_verifier' and provide the value below.")
    oauth_verifier = input('oauth_verifier: ')
    a.set_verifier(oauth_verifier)
    a.access_token.key 
    a.access_token.secret
    print("add the following to your .env")
    print(f"FLICKR_API_KEY={api_key}")
    print(f"FLICKR_API_SECRET={api_secret}")
    print(f"FLICKR_ACCESS_TOKEN={a.access_token.key}")
    print(f"FLICKR_ACCESS_SECRET={a.access_token.secret}")



import pydantic
import re


class DBModelMetaClass(pydantic.main.ModelMetaclass):
    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)
        cls.Config.arbitrary_types_allowed = True  # but, be careful if you use them!
        if getattr(cls.Config, "table_name", None):
            cls.__db_name__ = cls.Config.table_name
        else:
            cls.__db_name__ = re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()
        cls._db = None

        for name, field in cls.__fields__.items():
            setattr(cls, name, field)
        return cls

    @property
    def __db__(cls):
        return cls.handle_db_property(cls)

    @staticmethod
    def handle_db_property(cls):
        """Return the database client with model collection"""
        pass
    

class BaseDataModel(pydantic.BaseModel):

    key: str = None
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.key = self.key or None    

    @classmethod 
    def db(cls):
        return cls.__db__

    @classmethod
    def get(cls, key):
        pass
    
    @classmethod
    def get_all(cls):
        pass
    
    @classmethod
    def delete_key(cls, key):
        pass

    @classmethod 
    def query(cls, statement):
        pass

    def update(self, update_data: dict):
        pass
    
    def save(self):
        pass
    
    def get_attribute_value(self, index=1):
        d = self.__dict__
        return list(d.values())[index]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastmvc',
 'fastmvc.extras',
 'fastmvc.models',
 'fastmvc.models.database',
 'fastmvc.templates.core',
 'fastmvc.templates.core.static_pages',
 'fastmvc.templates.scaffold',
 'fastmvc.templates.users']

package_data = \
{'': ['*'],
 'fastmvc': ['templates/core/ignore/*',
             'templates/core/static_pages/static/css/*',
             'templates/core/static_pages/static/img/*',
             'templates/core/static_pages/static/js/*',
             'templates/core/static_pages/templates/*',
             'templates/scaffold/templates/*',
             'templates/scaffold_helpers/*',
             'templates/users/templates/*']}

install_requires = \
['Jinja2',
 'authlib',
 'fastapi',
 'httpx',
 'itsdangerous',
 'passlib[bcrypt]',
 'pyjwt',
 'python-dotenv',
 'python-multipart',
 'typer[all]']

entry_points = \
{'console_scripts': ['fastmvc = fastmvc.main:app']}

setup_kwargs = {
    'name': 'fastmvc',
    'version': '0.1.1',
    'description': 'Rapid application development built for the cloud.',
    'long_description': '# FastMVC\nFastMVC is a modern, fast (high-performance), web framework for building Web Applications\nwith the MVC structure (Model - View - Controller) and effortlessly deploying them to cloud platforms. \n\n- Model is interchangeable depending on the cloud platform you would like to use.\n- View uses Jinja2 to create front end pages\n- Controller is written using FastAPI\n\n\n## FastMVC CLI\n`fastmvc new [PROJECT_NAME]`  \nCreates a new project. Will ask which platform to build towards (GOOGLE_APP_ENGINE, or DETA) and set up the base of the project accordingly.  \n\n`fastmvc scaffold [MODEL_NAME] [ATTRIBUTE]:[DATA_TYPE]`  \nScaffold out a Model, View, and Controller for your object. For example:  \n\nfastmvc scaffold item title:str description:wysiwyg amount:int available:bool  \n\n`fastmvc auth`  \nBuilds an Authentication Framework to easily integrate user sign in for your application.  \n\n`fastmvc s`  \nAlias for `uvicorn main:app --reload` to run your application locally  \n\n## Supported Cloud Platforms\n__Built__\n- Google App Engine (using Firestore database)\n- Deta (using DetaBase)\n\n__Coming Soon__\n- AWS Elastic Beanstalk\n\n\n',
    'author': 'MBeebe',
    'author_email': 'pyn-sol@beebe.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



from deta import Deta
from fastmvc.models.database._base import BaseDataModel, DBModelMetaClass


class DetaModelMeta(DBModelMetaClass):

    @staticmethod
    def handle_db_property(cls):
        if cls._db:
            return cls._db
        cls._db = Deta().Base(cls.__db_name__)
        return cls._db
    

class DetaBase(BaseDataModel, metaclass=DetaModelMeta):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def get(cls, key):
        item = cls.__db__.get(key)
        return cls.parse_obj(item)
    
    @classmethod
    def get_all(cls):
        for record in cls.__db__.fetch().items:
            yield cls.parse_obj(record)
    
    @classmethod
    def delete_key(cls, key):
        cls.__db__.delete(key)

    @classmethod
    def query(cls, statement):
        for record in cls.__db__.fetch(statement).items:
            yield cls.parse_obj(record)

    def update(self, update_data: dict):
        for k, v in update_data.items():
            if k in self.__dict__:
                setattr(self, k, v)
        self.save()
    
    @classmethod
    def _db_put(cls, data, expire_in, expire_at):
        return cls.__db__.put(data, expire_in=expire_in, expire_at=expire_at)
    
    def save(self, expire_in: int or None = None, expire_at: int or None = None):
        """Saves the record to the database. Behaves as upsert, will create
        if not present. Database key will then be set on the object."""
        saved = self._db_put(self.dict(), expire_in, expire_at)
        self.key = saved["key"]



from authlib.integrations.starlette_client import OAuthError
from fastmvc.models.html import Alert
from fastapi import APIRouter, Request
from fastapi.responses import RedirectResponse, HTMLResponse
from fastapi.templating import Jinja2Templates

from user.model import User
from user.utils import current_user, authenticated_path, Auth


user_router = APIRouter()
templates = Jinja2Templates(directory="")


# SIGNUP
@user_router.get('/signup')
def signup_form(request: Request):
    return templates.TemplateResponse(
        'user/templates/signup.html',
        context={'request': request })


@user_router.post('/signup')
async def signup(request: Request):
    form_data = await request.form()
    user = User.parse_obj(form_data)
    response = user.signup()
    return templates.TemplateResponse(
        'user/templates/login.html',
        context={'request': request, 'alert': response})


#LOGIN
@user_router.get('/login')
def login_form(request: Request):
    continue_with_google = Auth.is_google_oauth2_enabled()
    return templates.TemplateResponse(
        'user/templates/login.html',
        context={'request': request, 'continue_with_google': continue_with_google})


@user_router.post('/login')
async def login(request: Request):
    form_data = await request.form()
    potential_user = User.parse_obj(form_data)
    result = potential_user.login(request)
    if isinstance(result, Alert):
        return templates.TemplateResponse(
        'user/templates/login.html',
        context={'request': request, 'alert': result})
    else:
        return result


@user_router.get('/dashboard')
@authenticated_path
def dashboard(request: Request):
    return templates.TemplateResponse(
        'user/templates/dashboard.html',
        context={'request': request, 'user': current_user(request) })


@user_router.get('/not_authenticated')
def not_authenticated(request: Request):
    return templates.TemplateResponse(
        'user/templates/not_authenticated.html',
        context={'request': request})


@user_router.get('/logout')
async def logout(request: Request):
    request.session.pop('user', None)
    return RedirectResponse(url='/')


@user_router.get('/refresh_token')
@authenticated_path
def refresh_token(request: Request):
    return {'access_token': User.refresh_token(request)}


@user_router.get('/username_available')
def username_available(u: str):
    return False if User.fetch_user(u) else True


# GOOGLE OAUTH
@user_router.get('/google_login')
async def login(request: Request):
    redirect_uri = request.url_for('auth')
    return await Auth.oauth.google.authorize_redirect(request, redirect_uri)


@user_router.get('/auth')
async def auth(request: Request):
    try:
        token = await Auth.oauth.google.authorize_access_token(request)
    except OAuthError as error:
        return HTMLResponse(f'<h1>{error.error}</h1>')
    user = token.get('userinfo')
    if user:
        db_user = User.signup_from_google_login(user.email)
        request.session['user'] = db_user.dict()
        request.session['user'].update(dict(user))

    return RedirectResponse(url='/user/dashboard')



from fastapi import APIRouter, Request
from fastapi.responses import RedirectResponse
from fastapi.templating import Jinja2Templates

from ^{obj}^.model import ^{Obj}^


^{obj}^_router = APIRouter()
templates = Jinja2Templates(directory="")


# INDEX
@^{obj}^_router.get('/')
def index(request: Request):
    ^{obj}^_list = ^{Obj}^.get_all()
    return templates.TemplateResponse(
        '^{obj}^/templates/index.html',
        context={'request': request, '^{obj}^_list': ^{obj}^_list })


# CREATE
@^{obj}^_router.get('/new')
def new(request: Request):
    return templates.TemplateResponse(
        '^{obj}^/templates/form.html',
        context={'request': request, 'vals': dict() })


@^{obj}^_router.post('/new', response_model=^{Obj}^)
async def create(request: Request):
    form_data = await request.form()
    ^{obj}^ = ^{Obj}^.parse_obj(form_data)
    ^{obj}^.save()
    return RedirectResponse(f'/^{obj}^/{^{obj}^.key}', status_code=303)


# UPDATE
@^{obj}^_router.get('/edit/{key}')
def edit(request: Request, key: str):
    vals = ^{Obj}^.get(key)
    return templates.TemplateResponse(
        '^{obj}^/templates/form.html',
        context={'request': request, 'vals': vals.dict() })


@^{obj}^_router.post("/edit/{key}")
async def update(request: Request, key: str):
    in_db = ^{Obj}^.get(key)
    update_data = await request.form()
    in_db.update(update_data)
    return RedirectResponse(f'/^{obj}^/{key}', status_code=303)


# DELETE
@^{obj}^_router.get('/delete/{key}')
def delete(request: Request, key: str):
    ^{Obj}^.delete_key(key)
    return RedirectResponse(f'/^{obj}^', status_code=303)


# VIEW
@^{obj}^_router.get('/{key}')
def view(request: Request, key: str):
    ^{obj}^ = ^{Obj}^.get(key)
    return templates.TemplateResponse(
        '^{obj}^/templates/view.html',
        context={'request': request, '^{obj}^': ^{obj}^ })
